var _ = require('lodash');
var V = require('mxd').Vector;
var R = require('mxd').Rect;
var T = require('ace/transform');

var p2 = require('p2');

var makeRunloop = require('ace/loop');
var makeEngine = require('ace/engine');
var graphics = require('ace/graphics');
var controls = require('ace/controls');
var makeDpad = require('ace/dpad');
var util = require('ace/util');

var music = require('./music');
var materials = require('./materials');
var items = require('./items');
var gfx = require('./gfx');
var vehicles = require('./vehicles');
var groups = require('./groups');
var sound = require('./sound');
var finale = require('./finale');

var runloop;
var gBg;



function makeBgLayer(ng, gx, depth, tile, scale) {
  let id = ng.makeID();

  var gi = graphics.image(gx, [0, 0]);
  gi.ord = 2;

  if (tile)
    gi = graphics.tile(gi);

  ng.systems.render.setGraphic(
    id,
    graphics.parallax(
      gi,
      depth
    )
  );

  ng.systems.transform.setTransform(id, T.scale(scale || [2, 2]));
  return id;
}

function makeBg(ng) {
  gBg = makeBgLayer(ng, gfx.bkg, [1,1], false, [1,1]);
}


function slope(ce, body) {
  var d = p2.vec2.dot(ce.normalA, [0, 1]); // Normal dot Y-axis
  if(ce.bodyB === body) d *= -1;
  return d;
}


function ground(engine, body) {
  var r = null;
  engine.systems.physics.world.narrowphase.contactEquations.forEach(function(ce) {
    if (ce.bodyA === body || ce.bodyB === body) {
      if (slope(ce, body) > .5)
        r = (ce.bodyA === body) ? ce.bodyB : ce.bodyA;
    }
  });
  return r;
}


function makePlayer(engine) {
  let shadow = engine.makeID();
  let player = engine.makeID();

  engine.systems.render.setGraphic(shadow, graphics.image(gfx.shadow));
  engine.addSystem({
    frame: function() {
      var tr = engine.systems.transform.getPos(player);
      engine.systems.transform.setPos(shadow, [tr[0], items.floor]);
    }
  });

  engine.systems.transform.setPos(player, [100, 400]);

  var body = new p2.Body({ mass: 1, position: [100, 400], fixedRotation: true });
  body.rgaKeep = true;

  var shape = new p2.Rectangle(35, 55);
  shape.collisionMask = groups.misc | groups.bounds;
  body.addShape(shape);
  shape.material = materials.player;

  var fs = new p2.Rectangle(35, 55);
  body.addShape(fs);
  fs.material = materials.foot;

  var standSensor = new p2.Circle(40);
  standSensor.sensor = true;
  body.addShape(standSensor, [0, 0]);

  var world = engine.systems.physics.world;

  var standItems = [];

  world.on('beginContact', function(e) {
    if (e.shapeA === standSensor || e.shapeB === standSensor) {
      let otherBody = (e.shapeA === standSensor) ? e.bodyB : e.bodyA;
      if (standItems.indexOf(otherBody.data) == -1) {
        if (otherBody.data !== undefined)
          standItems.push(otherBody.data);
      }
    }
  });

  world.on('endContact', function(e) {
    if (e.shapeA === standSensor || e.shapeB === standSensor) {
      let otherBody = (e.shapeA === standSensor) ? e.bodyB : e.bodyA;

      _.pull(standItems, otherBody.data);
    }
  });

  var cc=0;
  function crush() {
    if (cc++)return;
    body.removeShape(shape);
    body.removeShape(fs);
    body.mass = 0;
    body.velocity = [0, 0];
    setTimeout(die, 500);
  }

  function die() {
    sound.play("dead");
    var id = engine.makeID();
    engine.systems.render.setGraphic(id, graphics.image(gfx.dead, [0, 0]));
    engine.systems.reaper.kill(player);
    window.addEventListener('keydown', function() {
      window.location.reload();
    });
  }

  world.on('preSolve', function(e) {

    var floor;
    e.contactEquations.forEach(function(ce) {

      if (ce.shapeA !== fs && ce.shapeB !== fs)
        return;
      if (slope(ce, body) < .5)
        ce.enabled = false;
      else {
        floor = (ce.shapeA === fs) ? ce.shapeB : ce.shapeA;
      }
    });

    e.frictionEquations.forEach(function(ce) {
      if (ce.shapeA !== fs && ce.shapeB !== fs)
        return;

      var other = (ce.shapeA === fs) ? ce.shapeB : ce.shapeA;
      if (other != floor)
        ce.enabled = false;
    });

  });

  engine.systems.physics.setBody(player, body);


  // Gfx
  var stand = graphics.image(gfx['boy-stand']);
  var frames = [
    graphics.image(gfx['boy-walk-1']),
    stand,
    graphics.image(gfx['boy-walk-2']),
    stand
  ];
  var walk = graphics.animation(frames, 0.5);
  engine.systems.render.setGraphic(
    player,
    stand
  );

  var jumpKey = controls.key('J');


  // Controls
  var dpad = makeDpad();
  dpad.watch(function(dir) {
    materials.playerEnv.surfaceVelocity = dir[0] * 200;
    
    if (dir[0] < 0) {
      engine.systems.render.setGraphic(player, walk);
      let tr = engine.systems.transform.getTransform(player);
      engine.systems.transform.setTransform(player, T.copy(tr, {scale: [-1,1]}));
    }

    if (dir[0] > 0) {
      engine.systems.render.setGraphic(player, walk);
      let tr = engine.systems.transform.getTransform(player);
      engine.systems.transform.setTransform(player, T.copy(tr, {scale: [1,1]}));
    }

    if (dir[0] == 0) {
      engine.systems.render.setGraphic(player, stand);
    }

    if (jumpKey.get()) {
      var ppn = engine.systems.physics.getBody(player);
      let g = ppn && ground(engine, ppn.body);
      if (g) {
        sound.play("jump");
        engine.systems.physics.getBody(player).body.applyForce( [0, -20000], [0, 0]);
      }
    }


  });

  function pickup(item) {

    var pnode = engine.systems.physics.getBody(item);
    var gnode = engine.systems.render.nodes[item];

    if (pnode.body.rgaSnow) {

      if (gnode && gnode.invisible)
        return;

      if (gnode)
        gnode.invisible = true;

      var sb = items.makeSnowball(engine, [0, 0], true);
      pickup(sb);
      return;
    }

    sound.play("pickup");

    pnode.body.angularVelocity = 0;
    engine.systems.items.pickup(player, item);
  }

  function throwItem(item) {

    engine.systems.items.drop(item);

    var pnode = engine.systems.physics.getBody(item);

    if (pnode && pnode.body.rgaTag === 'apple') {
      engine.systems.reaper.kill(item);
      sound.play("windup");
      music.stop();
      setTimeout(function() { finale.reveal(engine, gBg, player); }, 2000);
      return;
    }

    sound.play("throw");

    var ptr = engine.systems.transform.getTransform(player);

    var vx = ptr.scale[0] > 0 ? 1 : -1;
    var vy = 0;
    vx *= 500;

    if (dpad.get()[1] < 0) {
      vx /= 3;
      vy = -400;
    }

    pnode.body.velocity = [vx, vy];
    pnode.body.angularVelocity = (Math.random() * 2 - 1) * 20;
  }

  controls.key('P').filter().watch(_.partial(finale.reveal, engine, gBg, player));

  controls.key('K').filter().watch(function() {

    var inv = engine.systems.items.holding(player);
    if (inv.length) {
      inv.forEach(function(item) {
        throwItem(item);
      });
    }
    else {
      var si = standItems.filter(function(s) {
        var gnode = engine.systems.render.nodes[s];
        return (gnode && !gnode.invisible);
      });
      if (si.length)
        pickup(si[0]);
    }
  });

  engine.addSystem({
    frame: function() {
      var transform = engine.systems.transform;
      var tr = transform.getTransform(player);
      var x = tr.pos[0];
      if (!(10 < x && x < 735)) 
        crush();

      if (tr.pos[1] > 450)
        die();
    }
  });

  return player;
}

function main() {
  music.play('assets/over.mid');
  gfx.loadGraphics().watch(function(g) {
    start();
  });
}

function start() {

  let engine = makeEngine({
    render: {
      width: 750,
      height: 500
    }
  });

  engine.addSystem({
    frame: function() {
      var transform = engine.systems.transform;
      _.forEach(transform.nodes, function(tr, id) {
        if (tr.pos[1] > 2000) {
          engine.systems.reaper.kill(id);
          console.log('kill OOB', id);
        }
      });
    }
  });



  engine.addSystem('items', items.system(engine));
  engine.addSystem('drive', vehicles.system(engine));

  engine.systems.physics.world.addContactMaterial(materials.playerEnv);
  engine.systems.physics.world.addContactMaterial(materials.envEnv);

  let bg = makeBg(engine);

  items.makePlatform(engine, R.fromSize([0, 440], [750, 60]));
  items.makePlatform(engine, R.fromSize([0, -60], [750, 60]));
  items.makePlatform(engine, R.fromSize([-60, 0], [60, 500]), groups.bounds);
  items.makePlatform(engine, R.fromSize([750, 0], [60, 500]), groups.bounds);

  items.makePlatform(engine, R.fromSize([60, 196], [260, 20]));
  items.makePlatform(engine, R.fromSize([496, 120], [130, 20]));

  let player = makePlayer(engine);

  items.makeApple(engine, [550, 100]);

  var zz = 0;

  function vl() {
    if (finale.going)
      return;
    var ty = util.choose([
      'pusher', 'minelayer', 'boxTruck', 'boxTruck', 'boxTruck'
    ]);
    if (!zz) {
      ty = 'boxTruck';
      zz = 1;
    }
    vehicles[ty](engine, [950, 400], [-100, 0], 1);
    setTimeout(vr, 12000);
  }
  
  function vr() {
    if (finale.going)
      return;
    var ty = util.choose(['pusher', 'pusher', 'pusher', 'minelayer', 'boxTruck']);
    vehicles[ty](engine, [-200, 400], [100, 0], -1);
    setTimeout(vl, 12000);
  }

  setTimeout(vl, 3000);

  var dbgId = engine.makeID();
  var dbg = engine.systems.physics.makeDebugGraphic();
  engine.systems.render.setGraphic(dbgId, dbg);

  runloop = makeRunloop({
    update: engine.update,
    frame: engine.frame
  });

  runloop.start();

  var title = engine.makeID();
  engine.systems.render.setGraphic(title, graphics.image(gfx.title));
  engine.systems.transform.setPos(title, [375, 250]);

  window.addEventListener('keydown', function(e) {
    if (title) {
      engine.systems.reaper.kill(title);
      title = null;
    }
    if (e.keyCode == 27)
      runloop.stop();
  });

}

window.addEventListener('load', main);
